/*
 * \file IMP/kernel_config.h
 * \brief Provide macros to mark functions and classes as exported
 *        from a DLL/.so, and to set up namespaces
 *
 * When building the module, IMP_EXPORTS should be defined, and when
 * using the module externally, it should not be. Classes and functions
 * defined in the module's headers should then be marked with
 * IMPEXPORT if they are intended to be part of the API, or with
 * IMPLOCAL if they are not (the latter is the default).
 *
 * The Windows build environment requires applications to mark exports in
 * this way; we use the same markings to set the visibility of ELF symbols
 * if we have compiler support.
 *
 * All code in this module should live in the IMP namespace.
 * This is simply achieved by wrapping things with the
 * IMP_BEGIN_NAMESPACE and IMP_END_NAMESPACE macros.
 * There are similar macros for module code that is designed to be for
 * internal use only.
 *
 * This header is auto-generated by tools/imp-module.py; it should not be
 * edited manually.
 *
 * Copyright 2007-2010 IMP Inventors. All rights reserved.
 *
 */

#ifndef IMP_CONFIG_H
#define IMP_CONFIG_H

#include <boost/static_assert.hpp>

#ifdef _MSC_VER
#ifdef IMP_EXPORTS
#define IMPEXPORT __declspec(dllexport)
#else
#define IMPEXPORT __declspec(dllimport)
#endif
#define IMPLOCAL
#else
#ifdef GCC_VISIBILITY
#define IMPEXPORT __attribute__ ((visibility("default")))
#define IMPLOCAL __attribute__ ((visibility("hidden")))
#else
#define IMPEXPORT
#define IMPLOCAL
#endif
#endif

#if _MSC_VER && !defined(SWIG) && !defined(IMP_DOXYGEN)
#ifdef IMP_EXPORTS

#define IMP_EXPORT_TEMPLATE(name)       template class __declspec(dllexport) name

#else

#define IMP_EXPORT_TEMPLATE(name)       template class __declspec(dllimport) name

#endif

#else
#define IMP_EXPORT_TEMPLATE(name)

#endif



#define IMP_BEGIN_NAMESPACE \
namespace IMP {\

#define IMP_END_NAMESPACE \
} /* namespace IMP */ \

#define IMP_BEGIN_INTERNAL_NAMESPACE \
IMP_BEGIN_NAMESPACE \
namespace internal {


#define IMP_END_INTERNAL_NAMESPACE \
} /* namespace internal */ \
IMP_END_NAMESPACE


#ifdef IMP_USE_CGAL
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#ifdef IMP_NO_CGAL
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_NO_CGAL
#ifdef IMP_NO_BOOST_LIBS
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#ifdef IMP_USE_BOOST_LIBS
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_USE_BOOST_LIBS
#ifdef IMP_NO_DEPRECATED
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#ifdef IMP_USE_DEPRECATED
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_USE_DEPRECATED
#ifdef IMP_LITTLE_ENDIAN
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_LITTLE_ENDIAN
#ifdef IMP_BOOST_VERSION
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_BOOST_VERSION 103301
#ifdef IMP_DEBUG
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_DEBUG 0
#ifdef IMP_RELEASE
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_RELEASE 1
#ifdef IMP_FAST
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_FAST 2
#ifdef IMP_BUILD
/* Do not define IMP config macros directly */
BOOST_STATIC_ASSERT(false)
#endif
#define IMP_BUILD IMP_RELEASE


//  functions are defined explicitly for swig
#ifndef SWIG
namespace IMP {
  class VersionInfo;
}

#include <IMP/internal/directories.h>
#include <IMP/kernel_config.h>
#include <string>

IMP_BEGIN_INTERNAL_NAMESPACE
IMPEXPORT std::string get_data_path(std::string module_name,
                                    std::string file_name);
IMPEXPORT std::string get_example_path(std::string module_name,
                                       std::string file_name);

IMP_END_INTERNAL_NAMESPACE

IMP_BEGIN_NAMESPACE
IMPEXPORT const VersionInfo& get_module_version_info();

inline std::string get_module_name() {
   return "IMP";
}

//! Return the path to installed data for this module
/** Each module has its own data directory, so be sure to use
    the version of this function in the correct module. To read
    the data file "data_library" that was placed in the \c data
    directory of module "mymodule", do something like
    \code
    std::ifstream in(IMP::mymodule::get_data_path("data_library"));
    \endcode
    This will ensure that the code works when \imp is installed or
    used via the \c tools/imppy.sh script.
*/
inline std::string get_data_path(std::string file_name) {
  return IMP::internal::get_data_path("kernel", file_name);
}

//! Return the path to installed example data for this module
/** Each module has its own example directory, so be sure to use
    the version of this function in the correct module.  For example
    to read the file \c example_protein.pdb located in the
    \c examples directory of the IMP::atom module, do
    \code
    IMP::atom::read_pdb(IMP::atom::get_example_path("example_protein.pdb", model));
    \endcode
    This will ensure that the code works when \imp is installed or
    used via the \c tools/imppy.sh script.
*/
inline std::string get_example_path(std::string file_name)  {
  return IMP::internal::get_example_path("kernel", file_name);
}


IMP_END_NAMESPACE

#endif // SWIG

#endif  /* IMP_CONFIG_H */
